part of fields.model;

typedef List<Field> CreateFields(Object, [Field parent]);

// TODO: are there any meaningful defaults we can provide?
final Map<Type, CreateFields> fieldListFactories = {};

// TODO: https://github.com/rmsmith/fields/issues/41
class FieldList extends ListBase<Field> {
  
  static final CreateFields defaultFactory = (Object o, [Field parent]) {
    if (fieldListFactories.containsKey(o.runtimeType)) {
      return fieldListFactories[o.runtimeType](o, parent);
    }
    return <Field>[];
  };
  
  static CreateFields factory = defaultFactory;
  
  static final Comparator defaultComparator = 
      (Field a, Field b) => a.displayName.compareTo(b.displayName);
  
  static final Function defaultPredicate = (Field f) => true;
    
  final List<Field> _source = new List<Field>();
  final List<Field> _view = new List<Field>();
  final List<StreamSubscription> _listeners = new List<StreamSubscription>();  
  final Field _parent;
    
  Comparator<Field> _comparator = defaultComparator;
  Comparator<Field> 
    get comparator => _comparator;
    set comparator(Comparator<Field> c) {
      assert(c != null);
      if (_comparator != c) {
        _comparator = c;
        for (Field f in _source) {
          f.children.comparator = c;
        }
        refreshView();
      }
    }
    
  final _onFieldChange = new StreamController<Field>();
  Stream<Field> get onFieldChange => _onFieldChange.stream;
  
  Function _predicate = defaultPredicate;
  Function
    get predicate => _predicate;
    set predicate(bool p(Field)) {
      assert(p != null);
      if (_predicate != p) {
        _predicate = p;
        for (Field f in _source) {
          f.children.predicate = p;
        }
        refreshView();
      }
    }
      
  int 
    get length => _view.length;
    set length(int length) => 
        throw new UnsupportedError('$FieldList is not modifiable.');
      
  FieldList([this._parent]);
  
  Field operator [](int index) => _view[index];
  
  void operator []=(int index, Field value) => 
      throw new UnsupportedError('$FieldList is not modifiable.');

  void refreshSource(Object object) {
    final fields = factory(object, _parent);
    // Remove any fields from the source list that are not in the new object.
    _source.removeWhere((f) => 
        // Consider fields equal if their name & type match.
        !fields.any((f2) => f2.name == f.name && f2.type == f.type));    
    fields.forEach((field) {
      // Consider fields equal if their name & type match.
      final predicate = (f) => f.name == field.name && f.type == field.type;
      if (_source.any(predicate)) {
        _source.singleWhere(predicate)
          .._getter = field._getter
          .._setter = field._setter
          ..refreshValue();
      }
      else {
        // Fields need to watch all of their siblings.
        fields
          .where((f) => f != field)
          .forEach((f) => f.onValueChange.listen((_) => field.refreshValue()));        
        _source.add(field);
      }
    });
    refreshView();
  }
      
  void refreshView() {    
    _listeners.forEach((s) => s.cancel());
    _listeners.clear();    
    _view
    ..clear()
    ..addAll(_source.where(predicate).toList()..sort(comparator))
    ..forEach((Field f) {
      _listeners.add(f.onValueChange.listen((_) {
        if (_onFieldChange.hasListener) {
          _onFieldChange.add(f);
        }
      }));
      f.children.refreshView();
    });
  }
}
